;;; The interface file for integrating the C KQML implementation with
;;; the LISP agents. This module provides the functions that can be called
;;; from the agent programs.

(in-package 'KQML)

(defpackage KQML
    (:export
     kqml-initialize        ;;; done
     kqml-register          ;;; done
     
     kqml-parse-buffer      ;;; done
     
     kqml-send              ;;; done
     kqml-build-msg         ;;; done
     kqml-free-msg          ;;; done
     
     kqml-put-field         ;;; done
     kqml-put-performative  ;;; done
     kqml-put-embedded-msg  ;;; done
     
     kqml-get-field         ;;; done
     kqml-get-ith-field     ;;; done
     kqml-get-performative  ;;; done
     kqml-get-field-count   ;;; done

     kqml-dbg-msg           ;;; temporarily
     
     kqml-print-msg         ;;; done
     kqml-msg-to-str))      ;;; done


;;; ***********************************************************************

;;; Local Global variables

(defvar *function-list* nil)

(defvar ANS-handler nil)


;;; ******** Local Functions ********

;;; Allocate space for a string

(defun malloc-foreign-string (str)
  (check-type str string)
  (let ((f-str (malloc-foreign-pointer :type
	       `(:pointer (:array :character (, (1+ (length str))))))))
    (setf (foreign-string-value f-str) str)
    (setf (foreign-pointer-type f-str) '(:pointer :character))
    f-str))

(def-foreign-function (C-null-pointer
		       (:return-type :pointer)
		       (:language :c)
		       (:name "_C_null_pointer")))


(defun convert-kmsg-to-list (kqml-mesg)
  (setf kmsg-list (list (__kqml_get_performative kqml-mesg)))
;  (format t "~a~%" kmsg-list)
  (setf fields (kqml_get_field_count kqml-mesg))
  (dotimes (i fields)
    (multiple-value-bind (field-value field-name value-len value-type)
	(__kqml_get_ith_field i kqml-mesg)
      (setf kmsg-list (append kmsg-list
		   (list (list field-name field-value value-len value-type))))))
;      (format t "~%~D : " i)
;      (format t "~a~%" kmsg-list)))
  kmsg-list)
			    

(defun convert-list-to-kmsg (list)
  (setf args (subseq list 4))
  (setf no-of-param (1+ (list-length args)))

  ;;; Allocate space for the parameters
  (setf fld-ptr (malloc-foreign-pointer :type
	       `(:pointer (:array (:pointer :character) (,no-of-param)))))
  (setf val-ptr (malloc-foreign-pointer :type
	       `(:pointer (:array (:pointer :character) (,no-of-param)))))
  (extract-field-value fld-ptr val-ptr args 0)
  (setf perf_ptr (malloc-foreign-string (symbol-name (first list))))
  (setf content_ptr (malloc-foreign-string (string (second (second list)))))
  (setf recv_ptr (malloc-foreign-string (string (second (third list)))))
  (setf reply_ptr (malloc-foreign-string (string (second (fourth list)))))

  ;;; Call the C function
  (setf kqml-mesg
    (kqml_build_msgv perf_ptr content_ptr recv_ptr reply_ptr fld-ptr val-ptr))
  
  ;;; Free the allocated space
  (free-foreign-pointer perf_ptr)
  (free-foreign-pointer content_ptr)
  (free-foreign-pointer recv_ptr)
  (free-foreign-pointer reply_ptr)
  (free-field-value fld-ptr val-ptr no-of-param)
  
  kqml-mesg)


(def-foreign-function kqml_build_msgv
    perf content receiver reply_with fld-ptr val-ptr)

;;; Functions called by convert-list-to-kmsg function

(defun extract-field-value (fld-ptr val-ptr args index)
  (when (not (endp args))
      (setf (foreign-aref fld-ptr index)
	     (malloc-foreign-string
	      (concatenate 'simple-string ":" (string (first (first args))))))
      (setf (foreign-aref val-ptr index)
	     (malloc-foreign-string (string (second (first args)))))
      (extract-field-value fld-ptr val-ptr (rest args) (+ index 1))))

(defun free-field-value (fld-ptr val-ptr no-of-param)
  (dotimes (i (1- no-of-param))
    (free-foreign-pointer (foreign-aref fld-ptr i))
    (free-foreign-pointer (foreign-aref val-ptr i)))
  (free-foreign-pointer fld-ptr)
  (free-foreign-pointer val-ptr))


(def-foreign-function kqml-free-msg kqml-mesg)

;;; kqml-get-field-count
(def-foreign-function (kqml_get_field_count
		       (:return-type :signed-32bit)
		       (:language :c)
		       (:name "_kqml_get_field_count"))
    kqml-mesg)

;;; kqml-get-performative
(defun __kqml_get_performative (kqml-mesg)
  ;;; Convert to symbol and return
  (intern (string-upcase
	   (foreign-string-value (kqml_get_performative kqml-mesg)))))

(def-foreign-function (kqml_get_performative
		       (:return-type (:pointer :character))
		       (:language :c)
		       (:name "_kqml_get_performative"))
    kqml-mesg)


;;; kqml-get-field

(defun __kqml_get_field (field kqml-mesg) 
  (setf field_ptr (malloc-foreign-string
		   (concatenate 'simple-string ":" (symbol-name field))))
  (setq value_len (malloc-foreign-pointer :type '(:pointer :signed-32bit)))
  (setq value_type (malloc-foreign-pointer :type '(:pointer :signed-32bit)))
  (setf field-value (foreign-string-value
	        (kqml_get_field field_ptr kqml-mesg value_len value_type)))
  (setf value-len (foreign-value value_len))
  (setf value-type (foreign-value value_type))
  (free-foreign-pointer field_ptr)
  (free-foreign-pointer value_len)
  (free-foreign-pointer value_type)
  (values field-value value-len value-type))

(def-foreign-function (kqml_get_field
		       (:return-type (:pointer :character))
		       (:language :c)
		       (:name "_kqml_get_field"))
    field kqml-mesg value-len value-type)

;;; kqml-get-ith-field
(defun __kqml_get_ith_field (field-no kqml-mesg)
  (setq field_name (malloc-foreign-pointer :type 
                      '(:pointer (:pointer :character))))
  (setq value_ptr (malloc-foreign-pointer :type
                      '(:pointer (:pointer :character))))
  (setq value_len (malloc-foreign-pointer :type '(:pointer :signed-32bit)))
  (setq value_type (malloc-foreign-pointer :type '(:pointer :signed-32bit)))
  (setf field-value (foreign-string-value 
     (kqml_get_ith_field field-no kqml-mesg field_name value_ptr 
                         value_len value_type)))
  ;;; The field name is set to be a symbol and returned
  (setf field-name (intern (string-upcase (subseq
		      (foreign-string-value (foreign-value field_name)) 1))))
  (setf value-len (foreign-value value_len))
  (setf value-type (foreign-value value_type))
  (free-foreign-pointer field_name)
  (free-foreign-pointer value_ptr)
  (free-foreign-pointer value_len)
  (free-foreign-pointer value_type)
  (values field-value field-name value-len value-type))

(def-foreign-function (kqml_get_ith_field
		       (:return-type (:pointer :character))
		       (:language :c)
		       (:name "_kqml_get_ith_field"))
  field-no kqml-mesg field-name field-value value-len value-type)


;;; **********************************************************************


;;; The front-end functions for all the KQML interface functions.


;;; Debug Functions

(defun kqml-dbg-msg (comment)
  (setf comment-ptr (malloc-foreign-string (string comment)))
  (C-kqml-dbg-msg comment-ptr))

(def-foreign-function (C-kqml-dbg-msg
		       (:language :c)
		       (:name "_C_kqml_dbg_msg"))
    (comment (:pointer :character)))
  

;;; ========================
;;; Initialization functions
;;; ========================

;;; kqml-initialize
(defun kqml-initialize (agent-name ans-handler-function)
    (setf agent_name (malloc-foreign-string (string agent-name)))
    (setf ANS-handler ans-handler-function)
    (setf result (C_kqml_initialize agent_name))
    (free-foreign-pointer agent_name)
    result)

(def-foreign-function (C_kqml_initialize
		       (:return-type :signed-32bit)
		       (:language :c)
		       (:name "_C_kqml_initialize"))
    (agent_name (:pointer :character)))




;;; ===============================
;;; Functions to register handlers.
;;; ===============================

;;; kqml-register

(defun kqml-register (function perf blocking)
  (kqml-dbg-msg "Creating char ptr for performative")
  (setf perf_ptr (malloc-foreign-string (symbol-name perf)))
  (kqml-dbg-msg "Adding new performative")
  (if (assoc perf *function-list*)
      (setf (second (assoc perf *function-list*)) function)
      (setf *function-list* (append (list (list perf function))
				    *function-list*)))
  (kqml-dbg-msg "Calling kqml-register")
  (setf result (kqml_register perf_ptr blocking))
  (free-foreign-pointer perf_ptr)
  result)


(def-foreign-function (kqml_register
		       (:return-type :signed-32bit)
		       (:language :c)
		       (:name "_C_kqml_register"))
    (perf (:pointer :character))
    (blocking :signed-32bit))
		       


;;; ================================
;;; Functions to build the messages.
;;; ================================


(defun kqml-build-msg (perf content receiver reply-with &rest args)
  (setf msg (append (list (intern (symbol-name perf)))
       (list (list 'content content (length (string content)) 0))
       (list (list 'receiver receiver (length (string receiver)) 0))
       (list (list 'reply-with reply-with (length (string reply-with))
		   0))))
  (loop while (not (endp args)) do
    (setf msg (append msg
		(list (list (intern (symbol-name (first args)))
			    (second args)
			    (length (string (second args)))
			    0))))
    (setf args (rest (rest args))))
;  (format t "In KQML-BUILD-MSG : The message~%")
;  (format t "~a~%" msg)
  msg)



;;; ===================================================
;;; Functions to access the fields of the KQML message.
;;; ===================================================

;;; kqml-get-field-count
(defun kqml-get-field-count (kqml-mesg)
  (length (rest kqml-mesg)))


;;; kqml-get-performative
(defun kqml-get-performative (kqml-mesg)
  (first kqml-mesg))


;;; kqml-get-field
(defun kqml-get-field (field kqml-mesg) 
  (let ((field-list (assoc (intern (symbol-name field)) (rest kqml-mesg))))
    (values (second field-list) (third field-list) (fourth field-list))))

;;; kqml-get-ith-field
(defun kqml-get-ith-field (field-no kqml-mesg)
  (values (first (first (subseq kqml-mesg (1+ field-no))))
	  (second (first (subseq kqml-mesg (1+ field-no))))
	  (third (first (subseq kqml-mesg (1+ field-no))))
	  (fourth (first (subseq kqml-mesg (1+ field-no))))))



;;; ================================================
;;; Functions to set the fields of the KQML message.
;;; ================================================


;;; kqml-put-performative

(defun kqml-put-performative (kqml-mesg perf)
  (if (consp kqml-mesg)
      (setf (first kqml-mesg) perf)
      (append (list perf) kqml-mesg))
  kqml-mesg)


;;; kqml-put-field

(defun kqml-put-field (field-name kqml-mesg value)
;  (format t "~% IN KQML-PUT-FIELD Function~%")
;  (format t "~% Field name : ~a~%" (intern (symbol-name field-name)))
  (if (assoc (intern (symbol-name field-name)) (rest kqml-mesg))
      (progn
;	(format t "Found ~a in the Message passed~%" field-name)
	(setf (second (assoc (intern (symbol-name field-name))
			     (rest kqml-mesg))) value)
	(setf (third (assoc (intern (symbol-name field-name))
			    (rest kqml-mesg)))
	      (length value)))
      (setf kqml-mesg (append kqml-mesg
			      (list (list (intern (symbol-name field-name))
					  value
					  (length (string value))
					  0)))))
  kqml-mesg)


;;; kqml-put-embedded-msg

(defun kqml-put-embedded-msg (field-name kqml-mesg embed-msg)
  (setf field-str (kqml-msg-to-str embed-msg))
;  (format t "~%Embedded mesg : ~S~%" field-str)
  (setf new-msg (kqml-put-field field-name kqml-mesg field-str))
  new-msg)



;;; ==========================================
;;; Function to Convert KQML message to string
;;; ==========================================

(defun kqml-msg-to-str (kqml-mesg)
  (setf kmsg (convert-list-to-kmsg kqml-mesg))
;  (format t "~% Converted List to KQML Message~%")
  (setf Str (foreign-string-value (kqml_msg_to_str kmsg)))
;  (format t "~%Converted KQML message to a string~%")
  (kqml-free-msg kmsg)
  Str)

(def-foreign-function (kqml_msg_to_str
		       (:return-type (:pointer :character))
		       (:language :c)
		       (:name "_kqml_msg_to_str"))
    kqml-mesg)



;;; ===============================
;;; Function to parse from a buffer
;;; ===============================

(defun kqml-parse-buffer (kmsg-string)
  (setf Str (malloc-foreign-string kmsg-string))
;  (format t "Created the String variable~%")
  (setf kmsg (C_kqml_parse Str (length kmsg-string)))
;  (format t "~%**** Created the KQML Message from the buffer ****~%")
  (setf kqml-mesg (convert-kmsg-to-list kmsg))
;  (format t "~% Converting KQML Message to a List~%")
  (free-foreign-pointer Str)
  (kqml-free-msg kmsg)
  kqml-mesg)

(def-foreign-function (C_kqml_parse
		       (:return-type :pointer)
		       (:language :c)
		       (:name "_C_kqml_parse"))
    Str len)



;;; ==============================
;;; Functions to send the messages
;;; ==============================

;;; kqml-send

(defun kqml-send (timeout kqml-mesg)
;  (format t "~%In kqml-send : Converting list to kqml_message~%")
  (setf kmsg (convert-list-to-kmsg kqml-mesg))
;  (format t "~%Converted the kqml message~%")
  (setf reply_msg (C-kqml-send timeout kmsg))
;  (format t "~%Returned from C-kqml-send~%")
  (if (zerop (foreign-pointer-address reply_msg))
      (free-foreign-pointer reply_msg)
      (progn
	(setf reply-mesg (convert-kmsg-to-list reply_msg))
;	(format t "~%******** IN KQML-SEND After Receiving the Reply *******~%")
	(kqml-print-msg reply-mesg)
	(kqml-free-msg kmsg)
	(kqml-free-msg reply_msg)
	reply-mesg)))

(def-foreign-function (C-kqml-send
		       (:return-type :pointer)
		       (:language :c)
		       (:name "_C_kqml_send"))
    timeout kqml-mesg)


;;; kqml-deliver

(def-foreign-function (kqml-deliver
		       (:return-type :signed-32bit)
		       (:language :c)
		       (:name "_kqml_deliver"))
    kqml-mesg)



;;; ===============================
;;; Functions to print the message.
;;; ===============================


(defun kqml-print-msg (kqml-mesg)
  (format t "~%~S~%" (first kqml-mesg))
  (setf kqml-mesg (rest kqml-mesg))
  (dolist (elem kqml-mesg)
	(format t "~a : " (first elem))
	(format t "~S " (second elem))
	(format t "~D " (third elem))
	(format t "~a~%" (fourth elem))))



;;; ===============
;;; Signal Handler.
;;; ===============

;;; The Handler function which is called from the C handler function
;;; This function checks the *function-list* for the performative and
;;; calls that function. Otherwise, returns an error.

(def-foreign-callable (lisp_handler (:return-type (:pointer :character)))
    ((content (:pointer :character))
     (kqml-mesg :pointer)
     (reply-type (:pointer :signed-32bit)))
  (setf perf (__kqml_get_performative kqml-mesg))
  (setf perf-func-pair (assoc perf *function-list*))
  (setf reply-mesg (C-null-pointer))
  (if perf-func-pair
      ;;; Call the function associated with the performative
      (progn
	(format t "~%Handler function : ~a~%" (cadr perf-func-pair))
	(setf list-of-kmsgs
	      (funcall (cadr perf-func-pair) (foreign-string-value content)
		       (convert-kmsg-to-list kqml-mesg)))
	(loop while (not (null list-of-kmsgs)) do
	      (setf reply-mesg
		    (C-kqml-build-kmsg-list reply-mesg
			     (convert-list-to-kmsg (first list-of-kmsgs))))
	      (format t "~% ******* KQML LIST OF MESSAGES ******** ~%")
	      (kqml-print-list reply-mesg)
	      (setf list-of-kmsgs (rest list-of-kmsgs)))
	(format t "~% Out of the while loop in the lisp-handler~%")
	(setf (foreign-pointer-type reply-mesg) '(:pointer :character))
	reply-mesg)
      
      ;;; Else call error handler
      (progn
	(setf reply-mesg (C-kqml-build-kmsg-list reply-mesg
			   (convert-list-to-kmsg (error-handler kqml-mesg))))
	(setf (foreign-pointer-type reply-mesg) '(:pointer :character))
	reply-mesg)
      ))

(defun error-handler (kqml-mesg)
  (setf kmsg (kqml-build-msg 'error "Not Ready for Performative"
				  (__kqml_get_field 'receiver kqml-mesg)
				  'nil)))

(def-foreign-function kqml-print-list kmsg-list)

(def-foreign-function (C-kqml-build-kmsg-list
		       (:return-type :pointer)
		       (:language :c)
		       (:name "_C_kqml_build_kmsg_list"))
    reply-mesg kqml-mesg)
